/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

 
package MetaTF.Parser;

import java.util.*;
import MetaTF.MetaTFConstants;
import MetaTF.Panic;

/**
 * The symbol table for the DTD
 *
 * The symbol table is a singleton
 *
 * @see <a href="metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.4 $, $Date: 2001/03/29 15:53:18 $),
 * @author Richard Jones
**/
public class DTD {
  private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Parser/DTD.java,v 1.4 2001/03/29 15:53:18 rej Exp $";

    /** The sections of the DTD: Map <section number, Section> **/
    private Map sections;
    
    /** The singleton implementation of the symbol table **/
    private static DTD impl = null;

    /**
     * All field names defined in the DTD: List <String>
     * Created by FieldLister, used by code generators when generating metadata API
     *
     * @see FieldLister
     * @see #addField
     * @see #addFields
     * @see MetaTF.Generator.Java.CodeGenerator
    **/
    private List allFields;

    /** Allocate a new singleton object representing the DTD **/
    private DTD() throws DTDException {
        sections = new HashMap();
        allFields = new LinkedList();
    }

    /**
     * The only way to get the symbol table
     *
     * @return the symbol table
    **/
    public static DTD instance() throws DTDException {
        if(impl == null) {
            impl = new DTD();
            impl.builtins();
	    }
        return impl;
    }

    /**
     * Does the DTD have a section of a given number
     *
     * @param n the section number
     * @return whether the DTD has a section n
    **/
    public boolean hasSection(int n) {
        return sections.containsKey(new Integer(n));
    }

    /**
     * Does the DTD have a section of a given name
     *
     * @param s the section name
     * @return whether the DTD has a section n
    **/
    public boolean hasSection(String s) {
        Iterator it = sections.keySet().iterator();
        while(it.hasNext()) {
            int n = ((Integer)it.next()).intValue();
            try {
                if(getSection(n).getName().equals(s))
                    return true;
            } catch(DTDException e) {
                Panic.panic("DTD.hasSection cannot get name of section "+n);
            }
        }
        return false;
    }

    /** @return the number of sections in the DTD **/
    public int getNumSections() {
        return sections.size();
    }

    /**
     * Get the number of the highest numbered section. Section numbers need not be
     * contiguous. This value is needed by code generators.
     *
     * @ return the number of the highest numbered section
     * @ see JCodeGenerator
    **/
    public int getMaxSectionNumber() {
        int max = 0;
        Iterator it = sections.keySet().iterator();
        while(it.hasNext()) {
            int n = ((Integer)it.next()).intValue();
            if(n > max)
                max = n;
        }
        return max;
    }
        
    /**
     * Return the section
     *
     * @param n the section number WHERE hasSection(n)
     * @return the section
     * @exception DTDException IF !hasSection(n)
     * @see #hasSection
    **/
    public Section getSection(int n) throws DTDException {
        if(!hasSection(n))
            throw new DTDException("No section "+n);
        else return (Section)sections.get(new Integer(n));
    }

    /** @return the sections **/
    public Map getSections() { return sections; }

    /**
     * Get section number from its name
     *
     * @param s the section name
     * @return the section number
     * @exception DTDException IF the DTD does not have section s
    **/
    public int sectionNameToNumber(String s) throws DTDException {
        Iterator it = sections.keySet().iterator();
        while(it.hasNext()) {
            int n = ((Integer)it.next()).intValue();
            if(getSection(n).getName().equals(s))
                return n;
        }
        throw new DTDException("DTD has no section "+s);
    }
        
    /**
     * Add a new section to the symbol table
     *
     * @param sectionNumber the section number
     * @param section the section
     * @exception DTDException IF sectionNumber<0 || hasSection(sectionNumber)
     * @see #hasSection
    **/
    void addSection(int sectionNumber, Section section)throws DTDException {
        if(sectionNumber < 0)
            throw new DTDException("l. "+section.getLine()+": Section numbers must be non-negative (found "+sectionNumber+")");
        if(hasSection(sectionNumber))
            throw new DTDException("l. "+section.getLine()+": Section "+sectionNumber+" already defined");
        sections.put(new Integer(sectionNumber), section);
    }

    /**
     * Construct the built-ins by hand
     * 1. Create a new section Global 0
     * 2. Add it to the symbol table
     * (from rel-1-2, we no longer create entries in the symbol table for buitl-in records)
    **/
    void builtins() throws DTDException {
        Section s0 = new Section("Global",0,false);
        addSection(0, s0);
    }

    /**
     * Append a field to the list of all named fields
     * Duplicate names are not added.
     *
     * @param o the name of the field
    **/
    void addField(String o) {
        if(!allFields.contains(o))
            allFields.add(o);
    }

    /**
     * Append a list of field to the list of all named fields.
     * Duplicate names are not added.
     *
     * @param fs the list of fields to be appended
    **/
    void addFields(List fs) {
        Iterator it = fs.iterator();
        while(it.hasNext()) {
            String o = (String)it.next();
            if(!allFields.contains(o))
                allFields.add(o);
        }
    }
    
    /**
     * Return a a list of all field names defined in the DTD: List <String>
     *
     * @return the list of field names
    **/
    public List getFields() { return allFields; }

    /**
     * Check and replace fields in a record
     * 1. foreach field name that refers to a record defined earlier,
     *        replace the NamedField with its definition
     * 2. foreach NamedField,
     *        check that it is well formed (has a type and an encoding, etc)
     *
     * @param r the record to be modified and checked
    **/
    private void checkAndReplace(Record r) throws DTDException {
        Section s0 = getSection(0);
        r.replaceFields(null, s0.getRecords(), new HashMap(s0.getFields()));
        r.checkFields();
    }

    /**
     * Dump the symbol table
     *
     * @param prefix an identifying prefix to prepend to each line of the dump
     * @return a prettified representation of the symbol table
    **/
    public String dump(String prefix) {
    	String result = "Symbol table\n";
    	Iterator it = sections.keySet().iterator();
    	while(it.hasNext()) {
    	    Integer n = (Integer)it.next();
    	    Section s = (Section)sections.get(n);
    	    result = result + s.dump(prefix+" ") + "\n";
    	}
    	return result;
    }

    /** Accept the visitor **/
    public Object accept(DTDVisitor visitor, Object data) {
        return visitor.visit(this, data);
    }

}

