/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

 
package MetaTF.Parser;

import java.util.*;
import MetaTF.Panic;

/**
 * Named field
 *
 * @see <a href="metatf.jjt">The MetaTF Grammar</a>
 * @see MetaTF.Parser.Field
 * @version rel-1-2 ($Revision: 1.4 $, $Date: 2001/03/29 15:53:18 $),
 * @author Richard Jones
**/
public class NamedField extends Field {
  private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Parser/NamedField.java,v 1.4 2001/03/29 15:53:18 rej Exp $";

    /** The name of the field **/
    private final String name;

    /** Its type **/
    private String type;

    /**
     * Its properties: Map <name, Value>
     *
     * @see MetaTF.Parser.Value
    **/
    private Map properties;

    /**
     * Allocate a new object representing a named field
     *
     * @param t the type
     * @param s the name of the field
     * @param p the property
     * @param v the value
     * @param l the line in the DTD at which this field was defined
     */
    NamedField(String t, String s, String p, Value v, int l) {
        type = t;
    	name = s;
    	properties = new HashMap();
    	if(p==null || p.equals(""))
    	    p = "encoding";
    	properties.put(p,v);
    	beginLine = l;
    }

    /**
     * Allocate a new object representing a named field
     *
     * @param t the type
     * @param s the name of the field
     * @param p the property
     * @param v the value
     */
    NamedField(String t, String s, String p, Value v) {
        this(t, s, p, v, 0);
    }

    /**
     * Allocate a new object representing a named field
     *
     * @param s the name of the field
     * @param l the line in the DTD at which this field was defined
     */
     NamedField(String s, int l) { 
    	type = null;
    	name = s;
    	properties = new HashMap();
    	beginLine = l;
    }

    /**
     * Allocate a new object representing a named field
     *
     * @param s the name of the field
     */
    NamedField(String s) {
        this(s, 0);
    }

    public String getName() { return name;}

    /**
     * @return a Map <name, Value> of the fields properties
     * @see MetaTF.Parser.Value
    **/
    public Map getProperties() { return properties; }

    /**
     * Does this field have a given property?
     *
     * @param name the name
     * @return whether the field has property name
    **/
    public boolean hasProperty(String name) {
        return properties.containsKey(name);
    }

    /**
     * @return a property
     *
     * @param name the name of the property
     * @return the value of the property
     * @see MetaTF.Parser.Value
    **/
    public Value getProperty(String name) {
        return (Value)properties.get(name);
    }

    Field replaceFields(Map inheritedRecords0, Map inheritedRecords, Map inheritedFields) {
    	if(!properties.isEmpty()) { // Some locally defined properties: inherit other properties
    	    inheritProperties(inheritedFields);
    	    return this;
    	}
    	else if(inheritedRecords!=null && inheritedRecords.containsKey(name)) {
    	    // Just a name defined - priority override is record in this section
    	    Record r = (Record)inheritedRecords.get(name);
    	    Field f = r.getField();
    	    if(f instanceof TupleField)
        		return f;
    	    else {
        		TupleField tfld = new TupleField(name);  // naming is a code generator issue
        		tfld.addField(f);
        		return tfld;
    	    }
    	}
    	else if(inheritedRecords0!=null && inheritedRecords0.containsKey(name)) {
    	    // Just a name defined - now try override with record from section 0
    	    Record r = (Record)inheritedRecords0.get(name);
    	    Field f = r.getField();
    	    if(f instanceof TupleField)
        		return f;
    	    else {
        		TupleField tfld = new TupleField(name);  // naming is a code generator issue
        		tfld.addField(f);
        		return tfld;
    	    }
    	}
    	else {      // Just a name defined - finally try inheriting fields
    	    inheritProperties(inheritedFields);
    	    return this;
    	}
    }

    /**
     * Add any inherited properties to this field
     * @param inherited fields a Map <name, NamedField> of inherited fields
     */
    private void inheritProperties(Map inheritedFields) {
    	if(inheritedFields==null) {
    	    return;
        }

        if(inheritedFields.containsKey(name)) {
            NamedField ifld = (NamedField)inheritedFields.get(name);
            if(properties.isEmpty()) 
                properties.putAll(ifld.properties);
            else {
                Map newprops = new HashMap(ifld.properties);
                newprops.putAll(this.properties);
                this.properties = newprops;
            }
            if(type==null || type.equals(""))
                type = ifld.type;
        }
    }


    /**
     * Check that this field has the following properties defined
     *  type, encoding
     * and that property is well-formed
     * All macro substition has been done (could check this with a flag?)
     */

    /**
     * Check that this field is well formed.
     * 1. It must have a type
     * 2. The only allowed property is 'encoding'
     * 3. The property must have a value
     * <PRE>
     *      type==string --> value is none or default <string>
     *      type==index  --> value is sectionOffset, sectionStride or default
     *      otherwise    --> the type may not be sectionOffset, sectionStride,
     *                       default <string> or a number
     * </PRE>
     * All macro substition must be done before checkFields is called
     *
     * @exception DTDException IF this field is not well-formed
     * @see #replaceFields
     */
    void checkFields() throws DTDException {

        if(type==null)
            error("type must be given for "+name+" (found null type)");
        if(type.equals(""))
            error("type must be given for "+name+" (found empty type)");
        if(!MetaTF.Types.knownType(type))
            error("unknown type "+type+" given for "+name);
             
        Iterator it = properties.keySet().iterator();
        boolean hasInterp = false;
        while(it.hasNext()) {
            String prop = (String)it.next();

            if(!prop.equals("encoding"))
                error("fields other than tag may only have property .encoding "+
                      "(found "+name+"."+prop+")");
            hasInterp = true;
            Value value = (Value)properties.get(prop);
            if(value==null)
                error("no encoding value given for "+name+"."+prop);
                
            if(value instanceof NumberValue)
                error("only tags may have values "+
                      "(found "+name+"."+prop+"="+value+")");

            else {
                try {
                    if(value instanceof SectionOffset)
                        ((SectionOffset)value).setNumber();
                    else if(value instanceof SectionStride)
                        ((SectionStride)value).setNumber();
                } catch (DTDException e) {
                    error(e.getMessage());
                }
            }

            if(type.equals("string")) {
                if(!((value instanceof None) ||
                     (value instanceof StringDefault)
                    ))
                    error("string fields may only have values None or Default "+
                          "(found string "+name+"."+prop+"="+value+")");

            } else if(type.equals("index")) {
                if(!((value instanceof SectionOffset) ||
                     (value instanceof SectionStride) ||
                     (value instanceof Default)
                    ))
                    error("index types may only have encoding sectionOffset, sectionStride or default 0 "+
                          "(found index "+name+"."+prop+"="+value+")");
                if(value instanceof Default) {
                    if(((Default)value).getDefault()!=0)
                        error("default value of index type must be 0 "+
                              "(found index "+name+"."+prop+"="+value+")");                       
                }

            } else {
                if((value instanceof StringDefault) ||
                   (value instanceof SectionOffset) ||
                   (value instanceof SectionStride) ||
                   (value instanceof NumberValue)
                  )
                    error("integral types cannot have encodings sectionOffset, sectionStride or string-valued default "+
                          "(found index "+name+"."+prop+"="+value+")");
            }
        }
        if(!hasInterp)
            error("encoding must be defined for "+name);
    }


    public boolean containsField(String s) {
        return name.equals(s);
    }

    /**
     * Add property p=v to this field
     */
    public void addProperty(String p, Value v, int l) throws DTDException {
        addProperty(type, name, p, v, l);
    }

    /**
     * Add a property to this field
     *
     * @param type the type
     * @param field the name of the field. The property is only added if the field names match
     * @param property the property (default 'encoding')
     * @param value the value of the property
     * @param beginLine the line in the DTD at which this property was defined
     * @exception DTDException IF
     * <PRE>
     *    1. this field already has this property defined
     *    2. the type is already set
     *    3. the type is string but the value is not default <string> or none
     *       the type is index but the value is not sectionOffset, sectionStride or default
     *       otherwise, if the value is sectionOffset, sectionStride, default <string>
     *       or a number
     * </PRE>
    **/
    public boolean addProperty(String type, String field, String property, Value value, int l) 
            throws DTDException {
    	if(!name.equals(field))
    	    return false;
         
        if(property==null || property.equals(""))
            property = "encoding";
        if(properties.containsKey(property))
            error("l. "+l+": "+property+"property already defined for "+name);
        if(this.type!=null)
	        error("l. "+l+": "+name+" already has type "+this.type);

	    if(type!=null)
    		this.type=type;    		
        else if(type.equals("string")) {
            if(!((value instanceof StringDefault) ||
                 (value instanceof None)
                ))
                error("l. "+l+": string types may only have encoding none or default <string> "+
                      "(found "+type+" "+field+"."+property+"="+value+")");
        } else if(type.equals("index")) {
            if(!((value instanceof SectionOffset) ||
                 (value instanceof SectionStride) ||
                 (value instanceof Default)
                ))
                error("l. "+l+": index types may only have encoding sectionOffset, sectionStride or default 0 "+
                      "(found "+type+" "+field+"."+property+"="+value+")");
            if(value instanceof Default) {
                if(((Default)value).getDefault()!=0)
                    error("l. "+l+": default value of index typemust be 0 "+
                          "(found "+type+" "+field+"."+property+"="+value+")");
            }
        } else {
            if((value instanceof StringDefault) ||
               (value instanceof SectionOffset) ||
               (value instanceof SectionStride) ||
               (value instanceof NumberValue)
              )
                error("l. "+l+": integral types cannot have encodings sectionOffset, sectionStride or string-valued default "+
                      "(found "+type+" "+field+"."+property+"="+value+")");
        }
                
        properties.put(property, value);
        return true;
    }

    //public only because it is an interface method
    public void addField(Field f) {
        Panic.panic("l. "+beginLine+": cannot add a field to a NamedField");
    }

    public int getNumNamedFields() { return 1; }

    /**
     * Get the tag of this field
     *
     * @exception NoTagException ALWAYS
    **/
    public int getTag() throws NoTagException {
        throw new NoTagException("l. "+beginLine+": NamedField");
    }

    public boolean isTagged() { return false; }

    public String getType() { return type; }

    /**
     * Get the type of the field if it matches the name given
     *
     * @param fname the name of the field whose type is required
     * @return the type IF fname matches ELSE null
    **/
    public String getType(String fname) {
        if(fname.equals(name))
            return type;
        else
            return null;
    }

    private void error(String s) throws DTDException {
        throw new DTDException(s+" (l. "+beginLine+")");
    }
        
    public String dump(String prefix) {
        String result = prefix;
        if(type==null)
            result += "null ";
        else
            result += type+" ";
        result += name+".";
        Iterator it = properties.keySet().iterator();
        while(it.hasNext()) {
            String prop = (String)it.next();
            result = result+prop+"="+((Value)properties.get(prop)).dump(prefix)+";";
        }
        return result;
    }

    public Object accept(DTDVisitor visitor, Object data) {
        return visitor.visit(this, data);
    }
}
