/*
    $File$
    Date        Author          Changes
    Nov 30 2000 Richard Jones   Created
    
    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */
 
package MetaTF.Runtime;

import MetaTF.MetaTFConstants;
import MetaTF.BadRecordException;
import MetaTF.Panic;
import java.io.*;

/**
 * Abstract superclass of all user-defined record classes                       
 *                                                                              
 * Record provides:                                                             
 * 1. write(TraceField, value) methods that understand interprations            
 * 2. putMetadata(writer, current section, tag, field, interpration [, values...]) methods 
 *    that change a field's encoding before emitting the corresponding    
 *    metadata record.                                                          
 *
 * @version rel-1-2 ($Revision: 1.2 $, $Date: 2001/03/29 15:53:24 $),
 * @author Richard Jones
**/
public abstract class Record {

    /**
     * Write this record
     *
     * @param llw the writer
     * @exception IOException IF the write fails
    **/
    public abstract void putRecord(LowLevelWriter llw) throws IOException;    


    /*                                                                                    
     * High-level write primitives.                                                        
     * These consult the TraceField for the field's encoding                         
     */                                                                                    
    protected void write(LowLevelWriter llw, NumTraceField tf, long v) throws IOException {                              
        switch(tf.getInterpretationType()) {                                               
            case MetaTFConstants.None:                                                     
                llw.write(v);                                                              
                break;                                                                     
            case MetaTFConstants.BaseOffset:                                               
                llw.write(v-tf.getValue());                                                
                break;                                                                     
            case MetaTFConstants.Delta:                                                    
                llw.write(v-tf.getValue());                                                
                try { tf.setValue(v); } catch(BadRecordException e) { /*not thrown */ }
                break;                                                                     
            case MetaTFConstants.Default:                                                  
            case MetaTFConstants.Stride:                                                   
                break;                                                                     
            case MetaTFConstants.SectionOffset:                                            
            case MetaTFConstants.SectionStride:                                            
                Panic.panic("section encoding type found in NumTraceField");             
        }                                                                                  
    }                                                                                      
                                                                                           
    protected void write(LowLevelWriter llw, IndexTraceField tf, int v) throws IOException {                             
        switch(tf.getInterpretationType()) {                                               
            case MetaTFConstants.Default:                                                  
            case MetaTFConstants.SectionStride:                                            
                break;                                                                     
            case MetaTFConstants.SectionOffset:                                            
                llw.write(v);                                                              
                break;                                                                     
            case MetaTFConstants.None:                                                     
            case MetaTFConstants.BaseOffset:                                               
            case MetaTFConstants.Delta:                                                    
            case MetaTFConstants.Stride:                                                   
                Panic.panic("indexed types may only have encodings Default, SectionOffset or SectionStride");             
        }                                                                                  
    }                                                                                      
                                                                                           
    protected void write(LowLevelWriter llw, StringTraceField tf, String v) throws IOException {                         
        switch(tf.getInterpretationType()) {                                               
            case MetaTFConstants.None:                                                     
                llw.write(v);                                                              
                break;                                                                     
            case MetaTFConstants.Default:                                                  
                break;                                                                     
            case MetaTFConstants.BaseOffset:                                               
            case MetaTFConstants.Delta:                                                    
            case MetaTFConstants.Stride:                                                   
            case MetaTFConstants.SectionOffset:                                            
            case MetaTFConstants.SectionStride:                                            
                Panic.panic("non-string encoding type found in StringTraceField");       
                System.exit(1);                                                            
        }                                                                                  
    }                                                                                      
                                                                                           


    /*                                                                                     
     * Metadata primitives.                                                                
     * These consult the TraceField for the field's encoding                         
     * These putMetadata methods are called by Record.putMetadata methods               
     * The typical user's interface is Record.putMetadata(...) which then calls         
     *     putMetadata(section, tag, ...)                                               
     */                                                                                  
    protected static void putMetadata(LowLevelWriter llw, int sectionNumber, int tag, int f, int i) throws BadRecordException, IOException {      
        TraceRecord tr;                                                                 
        TraceField tf;                                                                  
        try {                                                                           
            tf = MetaTFState.getTraceField(sectionNumber,tag,f);                                 
        } catch (ArrayIndexOutOfBoundsException e) {                                    
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        } catch (NullPointerException e) {                                              
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        }                                                                               
        tf.setInterpretationType(i);                                                     
        switch(i) {                                                                  
            case MetaTFConstants.None:                                               
                break;                                                               
            case MetaTFConstants.BaseOffset:                                         
            case MetaTFConstants.Delta:                                              
            case MetaTFConstants.Default:                                            
            case MetaTFConstants.SectionOffset:                                      
            case MetaTFConstants.Stride:                                             
            case MetaTFConstants.SectionStride:                                      
                throw new BadRecordException("Too few arguments given to putMetadata (tag="+tag+")"); 
            default:                                                                 
                throw new BadRecordException("putMetadata (tag="+tag+"): bad encoding "+i);     
        }                                                                            
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(tag);                                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.writeln();                                                                 
    }                                                                                  

    protected static void putMetadata(LowLevelWriter llw, int sectionNumber, int tag, int f, int i, long v) throws BadRecordException, IOException { 
        TraceRecord tr;                                                                 
        TraceField tf;                                                                  
        try {                                                                           
            tf = MetaTFState.getTraceField(sectionNumber,tag,f);                                 
        } catch (ArrayIndexOutOfBoundsException e) {                                    
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        } catch (NullPointerException e) {                                              
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        }                                                                               
        tf.setInterpretationType(i);                                                     
        switch(i) {                                                                  
            case MetaTFConstants.None:                                               
                throw new BadRecordException("Too many arguments given to putMetadata (tag="+tag+")"); 
            case MetaTFConstants.BaseOffset:                                         
            case MetaTFConstants.Delta:                                              
            case MetaTFConstants.Default:                                            
            case MetaTFConstants.SectionOffset:                                      
                tf.setValue(v);                                                      
                break;                                                               
            case MetaTFConstants.Stride:                                             
            case MetaTFConstants.SectionStride:                                      
                throw new BadRecordException("putMetadata (tag="+tag+"): stride and Section stride require 2 arguments"); 
            default:                                                                 
                throw new BadRecordException("putMetadata (tag="+tag+"): bad encoding "+i); 
        }                                                                            
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(tag);                                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v);                                                                  
        llw.writeln();                                                                 
    }                                                                                  

    protected static void putMetadata(LowLevelWriter llw, int sectionNumber, int tag, int f, int i, long v1, long v2) throws BadRecordException, IOException { 
        TraceField tf;                                                                  
        try {                                                                           
            tf = MetaTFState.getTraceField(sectionNumber,tag,f);                                 
        } catch (ArrayIndexOutOfBoundsException e) {                                    
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        } catch (NullPointerException e) {                                              
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        }                                                                               
        tf.setInterpretationType(i);                                                     
        switch(i) {                                                                  
            case MetaTFConstants.None:                                               
            case MetaTFConstants.BaseOffset:                                         
            case MetaTFConstants.Delta:                                              
            case MetaTFConstants.Default:                                            
            case MetaTFConstants.SectionOffset:                                      
                throw new BadRecordException("Too many arguments given to putMetadata (tag="+tag+")"); 
            case MetaTFConstants.Stride:                                             
            case MetaTFConstants.SectionStride:                                      
                tf.setValue(v1);                                                     
                   tf.setValue2(v2);                                                 
                   break;                                                            
            default:                                                                 
                throw new BadRecordException("Bad encoding (tag="+tag+") "+i); 
        }                                                                            
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(tag);                                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v1);                                                                 
        llw.write(v2);                                                                 
        llw.writeln();                                                                 
    }                                                                                  

    protected static void putMetadata(LowLevelWriter llw, int sectionNumber, int tag, int f, int i, String v) throws BadRecordException, IOException { 
        TraceField tf;                                                                  
        try {                                                                           
            tf = MetaTFState.getTraceField(sectionNumber,tag,f);                                 
        } catch (ArrayIndexOutOfBoundsException e) {                                    
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        } catch (NullPointerException e) {                                              
            throw new BadRecordException("Bad metadata record (section "+sectionNumber+", record "+tag+", field "+f); 
        }                                                                               
        tf.setInterpretationType(i);                                                       
        switch(i) {                                                                    
            case MetaTFConstants.None:                                                 
                throw new BadRecordException("Too many arguments given to putMetadata (tag="+tag+")"); 
            case MetaTFConstants.Default:                                              
                tf.setValue(v);                                                        
                break;                                                                 
            case MetaTFConstants.BaseOffset:                                           
            case MetaTFConstants.Delta:                                                
            case MetaTFConstants.SectionOffset:                                        
            case MetaTFConstants.Stride:                                               
            case MetaTFConstants.SectionStride:                                        
                throw new BadRecordException("Bad type for encoding"+i+" (tag="+tag+")"); 
            default:                                                                   
                throw new BadRecordException("Bad encoding "+i+" (tag="+tag+")");  
        }                                                                              
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(tag);                                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v);                                                                  
        llw.writeln();                                                                 
    }                                                                                  

} // end of class Record

